/*
 * Copyright (C) 2005 Andre Noll <maan@tuebingen.mpg.de>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file recv.c the stand-alone audio stream receiver */

#include <regex.h>
#include <sys/types.h>

#include "para.h"
#include "list.h"
#include "sched.h"
#include "ggo.h"
#include "buffer_tree.h"
#include "recv.h"
#include "recv.cmdline.h"
#include "fd.h"
#include "string.h"
#include "error.h"
#include "stdout.h"
#include "version.h"

extern void afh_recv_init(struct receiver *r);
#undef AFH_RECEIVER
#define AFH_RECEIVER {.name = "afh", .init = afh_recv_init},
DEFINE_RECEIVER_ARRAY;

/** The gengetopt args info struct. */
static struct recv_args_info conf;

static int loglevel;
/** Always log to stderr. */
INIT_STDERR_LOGGING(loglevel);

/** init array of error codes used by para_recv */
INIT_RECV_ERRLISTS;

__noreturn static void print_help_and_die(void)
{
	struct ggo_help h = DEFINE_GGO_HELP(recv);
	bool d = conf.detailed_help_given;

	ggo_print_help(&h, d? GPH_STANDARD_FLAGS_DETAILED : GPH_STANDARD_FLAGS);
	print_receiver_helps(d? GPH_MODULE_FLAGS_DETAILED : GPH_MODULE_FLAGS);
	exit(0);
}

/**
 * The main function of para_recv.
 *
 * \param argc number of arguments
 * \param argv vector of arguments
 *
 * para_recv uses the specified receiver to receive an audio stream sent by
 * para_server. The received data is written to stdout.
 *
 * \return \a EXIT_SUCCESS on success, \a EXIT_FAILURE on errors.
 */
int main(int argc, char *argv[])
{
	int ret, r_opened = 0, receiver_num;
	struct receiver *r = NULL;
	struct receiver_node rn;
	struct stdout_task sot = {.btrn = NULL};
	static struct sched s;
	struct task_info ti;

	recv_cmdline_parser(argc, argv, &conf);
	loglevel = get_loglevel_by_name(conf.loglevel_arg);
	version_handle_flag("recv", conf.version_given);
	recv_init();
	if (conf.help_given || conf.detailed_help_given)
		print_help_and_die();

	memset(&rn, 0, sizeof(struct receiver_node));
	rn.conf = check_receiver_arg(conf.receiver_arg, &receiver_num);
	if (!rn.conf) {
		PARA_EMERG_LOG("invalid receiver specifier\n");
		ret = -E_RECV_SYNTAX;
		goto out;
	}
	r = &receivers[receiver_num];
	rn.receiver = r;
	rn.btrn = btr_new_node(&(struct btr_node_description)
		EMBRACE(.name = r->name));
	ret = r->open(&rn);
	if (ret < 0)
		goto out;
	r_opened = 1;

	sot.btrn = btr_new_node(&(struct btr_node_description)
		EMBRACE(.parent = rn.btrn, .name = "stdout"));
	stdout_task_register(&sot, &s);

	ti.name = r->name;
	ti.pre_select = r->pre_select;
	ti.post_select = r->post_select;
	ti.context = &rn;
	rn.task = task_register(&ti, &s);

	s.default_timeout.tv_sec = 1;
	s.default_timeout.tv_usec = 0;
	ret = schedule(&s);
	sched_shutdown(&s);
out:
	if (r_opened)
		r->close(&rn);
	btr_remove_node(&rn.btrn);
	btr_remove_node(&sot.btrn);
	if (rn.conf)
		r->free_config(rn.conf);

	if (ret < 0)
		PARA_ERROR_LOG("%s\n", para_strerror(-ret));
	return ret < 0? EXIT_FAILURE : EXIT_SUCCESS;
}
